import os
import pymongo
import json
import re
import time
import datetime
import multiprocessing as mp
from math import ceil
from typing import List
from tweet_manager.lib import json2csv, format_csv

from config import globals

# Logger
import logging
logger = logging.getLogger(__name__)


def filesystem_writer(
        queue: mp.Queue,
        header: str,
        host: str,
        port: int,
        database: str,
        pagesize: int,
        output_dir: str,
        recovery_file: str)\
        -> None:
    """
    Reads the CSV pages from the queue and writes them to filesystem

    This function adds tweets to the files by appending them at the
    end. However, if the file or the dir structure didn't already exist,
    it will be created anew. The execution loop will exit only when a
    termination message is received.

    :param queue: queue were processed data await to be written to FS
    :param header: header of the new CSV files
    """

    logger.debug(
        "Worker {} launched: filesystem_writer executing"
        .format(os.getpid()))
    if recovery_file:
        recovery_file_path = recovery_file
    else:
        recovery_file_path = build_recovery_filepath(database)
        create_recovery_file(
            recovery_file_path, host, port, database, pagesize)

    while True:
        page_number, csv_page = queue.get()

        if csv_page == "END":
            logger.info("Writing loop finished")
            os.remove(recovery_file_path)
            break
        elif csv_page == "ERROR":
            logger.error("Dumping recovery file and exiting")
            if recovery_file_path >= 0:
                dump_error_recovery_file(recovery_file_path, page_number)
            break
        elif csv_page is not None:
            if globals.timing:
                time0 = time.time()
            for output_path in csv_page.keys():
                logger.debug("Dumping tweets for " + output_path)
                if os.path.isfile(output_path):
                    with open(output_path, "a") as writer:
                        writer.write(csv_page[output_path])
                else:
                    logger.debug("File {} not found, generating new..."
                                 .format(output_path))
                    generate_path(output_path, header)
                    with open(output_path, "a") as writer:
                        writer.write(csv_page[output_path])
            if page_number >= 0:
                update_recovery_file(recovery_file_path, page_number)
            if globals.timing:
                logger.critical(
                    "Time spent writing tweet page to FS: {}s"
                    .format(time.time() - time0))
            else:
                continue


def process_page(
        host: str,
        port: int,
        database: str,
        header: str,
        output_dir: str,
        pagesize: int,
        page_number: List[int],
        queue: mp.Queue)\
        -> None:
    """
    Retrieve a page of tweets from Mongo, convert to CSV and enqueue

    :param host: IP address of the MongoDB host
    :param port: Mongo port
    :param database: name of the database that we want to query to
    :param header: CSV header and all its fields
    :param output_dir: root directory for the new CSV database
    :param pagesize: size of page (number of tweets we want to get)
    :param page_number: number of the tweet page to retrieve
    :param queue: queue were processed data await to be written to FS
    """

    try:
        logger.debug(
            "Worker {} launched: process_page executing".format(os.getpid()))
        if globals.timing:
            time0 = time.time()

        client = pymongo.MongoClient(host, port)
        database_tweets = client[database]["tweets"]
        tweets_page = get_tweets_page(database_tweets, pagesize, page_number)
        buffer_tweets = {}

        for tweet in tweets_page:
            csv_tweet_output_path =\
                get_tweet_output_path(tweet, output_dir)
            try:
                csv_tweet_contents =\
                    "\n" + str(convert_tweet_to_csv(header, tweet))
            except TweetConversionException as exc:
                logger.error(exc.message)
                logger.error("Origin tweet:\n" + str(exc.tweet))
                logger.error("Discarding tweet and proceeding...")
                continue
            if csv_tweet_output_path not in buffer_tweets.keys():
                buffer_tweets[csv_tweet_output_path] = ""
            buffer_tweets[csv_tweet_output_path] += csv_tweet_contents
        client.close()

        queue.put((page_number, buffer_tweets))

        logger.debug("Page {} enqueued".format(page_number))
        if globals.timing:
            time1 = time.time()
            logger.critical(
                "Time processing & buffering tweet page: {}s"
                .format(time1 - time0))

    except Exception:
        raise ExceptionAtPage(
            "Something failed while processing page", page_number)


def get_tweet_output_path(tweet: dict, output_dir: str) -> str:
    """
    Generate a filesystem path string for the new CSV using tweet date

    :param tweet: tweet object in its Mongo JSON/dict form
    :param output_dir: root directory for the new CSV database
    :returns: path of the CSV file where the tweet should be written to
    """

    collection_path = output_dir

    # Extract year, month and date from the tweet using a regex
    matchObj = re.search(
        r"^(\d{4})-(\d{2})-(\d{2}) \d{2}:\d{2}:\d{2}$",
        str(tweet["created_at"]))
    year = matchObj.group(1)
    month = matchObj.group(2)
    day = matchObj.group(3)

    tweet_output_file = os.path.join(
        collection_path, year, month, day + ".csv")

    return tweet_output_file


def generate_path(path: str, header: str) -> None:
    """
    Generate dir structure and output file if they didn't exist yet

    :param path: path of the output file
    :param header: header of the new CSV files
    """

    path_regex = r"(\./\w+)/(\w+)/(\d{4})/(\d{2})/(\d{2})\.csv$"
    match = re.search(path_regex, path)

    output_dir = match.group(1)
    collection = match.group(2)
    year = match.group(3)
    month = match.group(4)
    day = match.group(5)

    if not os.path.isdir(output_dir):
        logger.info(
            "Building directory to contain the collected tweets at: "
            + os.path.abspath(output_dir)
        )
        os.mkdir(output_dir)

    collection_path = os.path.join(output_dir, collection)
    if not os.path.isdir(collection_path):
        logger.info("Initializing collection " + collection + "...")
        os.mkdir(collection_path)

    collection_year = os.path.join(collection_path, year)
    if not os.path.isdir(collection_year):
        logger.info("Adding year " + year + "...")
        os.mkdir(collection_year)

    collection_month = os.path.join(collection_year, month)
    if not os.path.isdir(collection_month):
        logger.info("Adding month " + month + "...")
        os.mkdir(collection_month)

    csv_output_file = os.path.join(collection_month, day + ".csv")
    if os.path.isfile(csv_output_file) is False:
        with open(csv_output_file, "w") as fw:
            logger.info("Adding file for day " + day + "...")
            fw.write(header.strip())


def convert_tweet_to_csv(header: str, tweet: dict) -> str:
    """
    Convert a JSON/dict-like tweet to a single string line in CSV form

    Takes the JSON/dict object passed as an argument and performs the
    following transformations:
    1. Flattening (removal of nested elements)
    2. Conversion to CSV (arg1: flat tweet, arg2: array compression,
       arg3: depth of array compression, arg4: remove dollars mode)
    3. Format the CSV to make it suitable to append to the file (by
       keeping just the information indicated in the header amongst
       other required manipulations)
    This function includes the core of the computations in the whole
    script, and is surely the most expensive part.
    For more information about how things get done and to see if
    performance might be enhanced, take a look at the tweet_manager
    library, which is probably an example of pretty bad code.

    :param header: header of the new CSV files
    :param tweet: tweet object in its Mongo JSON/dict form
    :returns: single line string with the tweet info for the header
              fields in CSV form
    """

    try:
        flat_tweet = json2csv.flatten_dictionary(tweet)
    except Exception:
        raise TweetConversionException(
            "Error when flattening tweet", tweet)
    try:
        csv_tweet = json2csv.json2csv(flat_tweet, True, 5, False)
    except Exception:
        raise TweetConversionException(
            "Error when trying to convert tweet to CSV", flat_tweet)
    try:
        csv_appendable_tweet = format_csv.get_csv_line(
            header, csv_tweet)
    except Exception:
        raise TweetConversionException(
            "Error when formatting CSV tweet", csv_tweet)
    return csv_appendable_tweet


def build_recovery_filepath(dbname: str) -> str:
    """
    Build the path of a recovery file

    :param dbname: name of the database being queried
    :returns: the path of the recovery file generated in this execution
    """
    recovery_dir = "./recovery"
    if not os.path.isdir(recovery_dir):
        os.mkdir(recovery_dir)
    now = datetime.datetime.now()
    datetime_str = "%04d%02d%02d-%02d%02d%02d" %\
        (now.year, now.month, now.day, now.hour, now.minute, now.second)
    recovery_file_path = os.path.join(
        recovery_dir,
        "recovery_" + dbname + "_" + datetime_str + ".json")

    return recovery_file_path


def create_recovery_file(
        file_path: str,
        host: str,
        port: int,
        database: str,
        page_size: int)\
        -> None:
    """
    In case of error, dump information to file to allow recovery

    :param host: address of the host to which the script connected
    :param port: port of the Mongo database
    :param database: name of the database being queried
    :param page_size: size of the page that was being used
    """

    recovery_file_contents = {}
    recovery_file_contents["host"] = host
    recovery_file_contents["port"] = port
    recovery_file_contents["database"] = database
    recovery_file_contents["pagesize"] = page_size
    recovery_file_contents["dumped_pages"] = []

    with open(file_path, "w") as f:
        json.dump(recovery_file_contents, f)

    logger.error("Generated recovery file at {}".format(file_path))


def update_recovery_file(
        file_path: str,
        page_number: int)\
        -> None:
    """
    Add a new page to the list of already dumped pages in the recovery
    file
    :param file_path: path to the recovery file
    :param page_number: number of the page that was safely written
    """

    with open(file_path, "r") as f:
        recovery_file_contents = json.load(f)
    recovery_file_contents["dumped_pages"].append(page_number)
    with open(file_path, "w") as f:
        json.dump(recovery_file_contents, f)


def dump_error_recovery_file(
        file_path: str,
        page_number: int)\
        -> None:
    """
    Add information pointing to the page where error was detected
    :param file_path: path to the recovery file
    :param page_number: number of the page that crashed
    """

    with open(file_path, "r") as f:
        recovery_file_contents = json.load(f)
    recovery_file_contents["error_page"] = page_number
    with open(file_path, "w") as f:
        json.dump(recovery_file_contents, f)


#########################
#  TWEET DB PAGINATION  #
#########################
def get_page_index(
        collection: pymongo.collection.Collection,
        page_size: int)\
        -> List[int]:
    """
    Get a list of the pages in the collection

    :param collection: pymongo collection of tweets
    :param page_size: number of tweets in each page
    :returns: list of pages in the collection, starting from 0
    """

    return list(range(0, ceil(collection.count() / page_size)))


def get_tweets_page(
        collection: pymongo.collection.Collection,
        page_size: int,
        num_page: int)\
        -> pymongo.cursor.Cursor:
    """
    Get a cursor pointing to the Mongo entries for that page

    :param collection: pymongo collection of tweets
    :param page_size: number of tweets in each page
    :param num_page: relative index of the page within the collection
    :returns: a Pymongo cursor pointing to the tweets
    """

    tweets =\
        collection.find().skip(num_page * page_size).limit(page_size)
    return tweets


#########################
#  METADATA GENERATION  #
#########################
def generate_metadata_file(collection_path: str) -> None:
    """
    Create a metadata file for the current collection

    Once all the CSV files have been created, generate a metadata file
    with information about the number of tweets in each of the CSVs by
    making a simple count of lines. Making a call to the system to
    perform "wc -l" would surely be faster, but more system dependant
    and a bit "hacky"

    :param collection_path: path to the root dir of a database
    """

    metadata = {}
    metadata["files"] = {}
    for root, dirs, files in os.walk(collection_path):
        for f in files:
            file_path = os.path.join(root, f)
            relative_path = os.path.relpath(file_path, collection_path)
            metadata["files"][relative_path] = {}
            metadata["files"][relative_path]["count"] =\
                file_length(file_path)
    output_path = os.path.join(collection_path, ".metadata.json")
    with open(output_path, 'w') as f:
        json.dump(metadata, f)


def file_length(file_path: str) -> int:
    """
    Calculate number of lines of a file

    :param file_path: path of the file
    :returns: number of lines of the file
    """

    with open(file_path) as f:
        for i, l in enumerate(f):
            pass
        return i


#######################
#  CUSTOM EXCEPTIONS  #
#######################
class ExceptionAtPage(Exception):
    """
    Exception designed to be raised when the conversion of a page of
    tweets taken from Mongo fails
    """

    def __init__(self, message: str, error_page: int):
        """
        :param message: str descriptive of the error
        :param error_page: int indicating the number of page that failed
        """

        self.message = message
        self.error_page = error_page


class TweetConversionException(Exception):
    """
    Should be raised when a tweet raises an exception in the process of
    being converted
    """

    def __init__(self, message: str, tweet: str):
        """
        :param message: str descriptive of the error
        :param tweet: str with the contents of the tweet that caused the
        failure
        """
        self.message = message
        self.tweet = tweet
